/*
 *  Impliments protocol specific functions for ICFP 2002 Programming Contest - Robot Game
 *
 *
 *
 */ 


#include <stdlib.h>
#include <string.h>

#include "debug.h"

#include "errors.h"
#include "network.h"
#include "protocol.h"


#define BUFFER_STEP_SIZE 1000


int readbuffer_size=0;
int readbuffer_offset=0;
int readbuffer_valid=0;
static char *readbuffer=NULL;


int proto_enlargebuffer();
int proto_receiveline();
int proto_freesendopslist(ll_oplist **slist);



int proto_sendhello(void)
{
	return proto_sendfullbuffer("Player\n",7);
}


int proto_receiveboard(mapdata *data)
{
	int rval;
	char *w,*h;
	int current_row;
	ll_rowdata *datablock;


	data->rows=NULL;

	rval=proto_receiveline();
	
	if (FAILURE(rval))
	{
		DEBUG_PRINT("Failed on proto_receiveline().\n");
		return rval;
	}


	/* Expecting two ASCII formated integers  width  height */
	w=strtok(readbuffer," ");
	h=strtok(NULL," ");

	if (w==NULL || h==NULL)
	{
		DEBUG_PRINT("Malformed server board dimension response.\n");
		return ERR_PROTO_MALFORMEDRESPONSE;
	}
	data->width=atoi(w);
	data->height=atoi(h);


	for (current_row=1;current_row<=data->height;current_row++)
	{
		datablock=malloc(sizeof(ll_rowdata));
		if (datablock==NULL)
			return ERR_OUTOFMEMORY;
		datablock->next=data->rows;
		datablock->rownumber=current_row;
		
		/* Get a map line */
		rval=proto_receiveline();
		
		if (FAILURE(rval))
		{
			DEBUG_PRINT("Failed on proto_receiveline().\n");
			return rval;
		}

		datablock->rowdata=strdup(readbuffer);

		if (datablock->rowdata==NULL)
			return ERR_OUTOFMEMORY;

		data->rows=datablock;
	}


	return 0;

}


int proto_receivestats(init_stats *data)
{
	int rval;
	char *pnum,*weight,*bid;

	rval=proto_receiveline();
	
	if (FAILURE(rval))
	{
		DEBUG_PRINT("Failed on proto_receiveline().\n");
		return rval;
	}

	pnum=strtok(readbuffer," ");
	weight=strtok(NULL," ");
	bid=strtok(NULL," ");

	data->playernum=atoi(pnum);
	data->maxweight=atoi(weight);
	data->bidleft=atoi(bid);

	return 0;

}


int proto_freepackagelist(ll_packagelist **plist)
{
	ll_packagelist *flist,*next;

	flist=*plist;
	while (flist!=NULL)
	{
		next=flist->next;
		free(flist);
		flist=next;
	}
	*plist=NULL;
	return 0;
}


int proto_receivepackages(ll_packagelist **data)
{
	int rval;
	char *pnum,*x,*y,*weight;
	ll_packagelist *head;
	ll_packagelist *newentry;

	rval=proto_receiveline();
	
	if (FAILURE(rval))
	{
		DEBUG_PRINT("Failed on proto_receiveline().\n");
		return rval;
	}

	if (*data!=NULL)
		proto_freepackagelist(data);


	head=NULL;

	pnum=strtok(readbuffer," ");

	while (pnum!=NULL)
	{
		x=strtok(NULL," ");
		y=strtok(NULL," ");
		weight=strtok(NULL," ");

		newentry=(ll_packagelist *)malloc(sizeof(ll_packagelist));
		if (newentry==NULL)
			return ERR_OUTOFMEMORY;

		newentry->packageid=atoi(pnum);
		newentry->x=atoi(x);
		newentry->y=atoi(y);
		newentry->weight=atoi(weight);

		newentry->next=head;
		head=newentry;

		pnum=strtok(NULL," ");
	}


	*data=head;

	return 0;

}

/*
 *  TODO: A command structure needs to be defined
 *
 *
 *
 */
int proto_sendcommand(sendcommand *cmd)
{
	int rval;
	static char outbuffer[32];
	

	switch(cmd->command)
	{
	case CMD_SND_MOVE:
		/* One operand, direction */
		if (cmd->ops==NULL)
		{
			DEBUG_PRINT("ERROR! Move command sent without direction!\n");
			/* Cover with an idle, and don't waste any bid on it */
			sprintf(outbuffer,"1 Pick\n");
			rval=proto_sendfullbuffer(outbuffer,strlen(outbuffer));
			break;
		}
		sprintf(outbuffer,"%d Move ",cmd->bid);
		rval=proto_sendfullbuffer(outbuffer,strlen(outbuffer));
		if (FAILURE(rval))
			return rval;
		switch(cmd->ops->op)
		{
			case DIRECTION_N:
				rval=proto_sendfullbuffer("N\n",2);
				break;
			case DIRECTION_S:
				rval=proto_sendfullbuffer("S\n",2);
				break;
			case DIRECTION_E:
				rval=proto_sendfullbuffer("E\n",2);
				break;
			case DIRECTION_W:
				rval=proto_sendfullbuffer("W\n",2);
				break;
			default:
				DEBUG_PRINT("WARNING! Default direction code hit!\n");
				rval=proto_sendfullbuffer("W\n",2);
				break;
		}

		break;
	case CMD_SND_PICK:
		{
			ll_oplist *c_op;

			sprintf(outbuffer,"%d Pick",cmd->bid);
			rval=proto_sendfullbuffer(outbuffer,strlen(outbuffer));
			if (FAILURE(rval))
				return rval;

			c_op=cmd->ops;
			while (c_op!=NULL)
			{
				sprintf(outbuffer," %d",c_op->op);
				rval=proto_sendfullbuffer(outbuffer,strlen(outbuffer));
				if (FAILURE(rval))
					return rval;
				c_op=c_op->next;
			}
			rval=proto_sendfullbuffer("\n",1);

		}
		break;
	case CMD_SND_DROP:
		{
			ll_oplist *c_op;

			sprintf(outbuffer,"%d Drop",cmd->bid);
			rval=proto_sendfullbuffer(outbuffer,strlen(outbuffer));
			if (FAILURE(rval))
				return rval;

			c_op=cmd->ops;
			while (c_op!=NULL)
			{
				sprintf(outbuffer," %d",c_op->op);
				rval=proto_sendfullbuffer(outbuffer,strlen(outbuffer));
				if (FAILURE(rval))
					return rval;
				c_op=c_op->next;
			}
			rval=proto_sendfullbuffer("\n",1);

		}
		break;

	default:
		/* Idle */
		sprintf(outbuffer,"1 Pick\n");
		rval=proto_sendfullbuffer(outbuffer,strlen(outbuffer));
		break;
	}

	if (FAILURE(rval))
		return rval;

	/* Free ops list */
	return proto_freesendopslist(&(cmd->ops));
}


int proto_freesendopslist(ll_oplist **slist)
{
	ll_oplist *flist,*next;

	flist=*slist;
	while (flist!=NULL)
	{
		next=flist->next;
		free(flist);
		flist=next;
	}
	*slist=NULL;
	return 0;
}

int proto_freereceivelist(ll_rcvlist **rlist)
{
	ll_rcvlist *flist,*next;

	flist=*rlist;
	while (flist!=NULL)
	{
		next=flist->next;
		free(flist);
		flist=next;
	}
	*rlist=NULL;
	return 0;
}



int proto_receivecommands(ll_rcvlist **rlist)
{
	int rval;
	int operating_player=-1;
	ll_rcvlist *head,*current;
	ll_rcvlist *newentry;


	char *token;

	if (*rlist!=NULL)
		proto_freereceivelist(rlist);

	head=current=NULL;

	rval=proto_receiveline();
	
	if (FAILURE(rval))
	{
		DEBUG_PRINT("Failed on proto_receiveline().\n");
		return rval;
	}
	

	/* Parse the line */
	token=strtok(readbuffer," ");
	while (token!=NULL)
	{
		switch(token[0])
		{
		case '#':
			/* Different player */

			operating_player=atoi(token+1);
			break;
		case 'D':
		case 'd':
			/* Drop */
			newentry=(ll_rcvlist *)malloc(sizeof(ll_rcvlist));
			if (newentry==NULL)
				return ERR_OUTOFMEMORY;
			newentry->playernum=operating_player;
			newentry->command=CMD_RCV_DROP;
			newentry->op2=0;

			token=strtok(NULL," ");
			newentry->op1=atoi(token);
			newentry->next=NULL;

			/* Add to end of list */
			if (current==NULL)
				head=current=newentry;
			else
			{
				current->next=newentry;
				current=newentry;
			}

			break;

		case 'P':
		case 'p':
			/* Pickup */
			newentry=(ll_rcvlist *)malloc(sizeof(ll_rcvlist));
			if (newentry==NULL)
				return ERR_OUTOFMEMORY;
			newentry->playernum=operating_player;
			newentry->command=CMD_RCV_PICK;
			newentry->op2=0;

			token=strtok(NULL," ");
			newentry->op1=atoi(token);
			newentry->next=NULL;

			/* Add to end of list */
			if (current==NULL)
				head=current=newentry;
			else
			{
				current->next=newentry;
				current=newentry;
			}
			break;


		case 'X':
		case 'x':
			/* X, Y location */
			newentry=(ll_rcvlist *)malloc(sizeof(ll_rcvlist));
			if (newentry==NULL)
				return ERR_OUTOFMEMORY;
			newentry->playernum=operating_player;
			newentry->command=CMD_RCV_POS;

			token=strtok(NULL," ");
			newentry->op1=atoi(token);
			token=strtok(NULL," ");
			if (strcmp(token,"Y"))
				DEBUG_PRINT("Warning! X token found without Y.\n");
			token=strtok(NULL," ");
			newentry->op2=atoi(token);
			newentry->next=NULL;

			/* Add to end of list */
			if (current==NULL)
				head=current=newentry;
			else
			{
				current->next=newentry;
				current=newentry;
			}

			break;

		default:
			/* Movement */
			if (token[0]!='N' && token[0]!='S' && token[0]!='E' && token[0]!='W')
			{
				sprintf(readbuffer,"Invalid token: %s",token);
				DEBUG_PRINT(readbuffer);
				return ERR_DATA_INVALIDTOKEN;
			}


			newentry=(ll_rcvlist *)malloc(sizeof(ll_rcvlist));
			if (newentry==NULL)
				return ERR_OUTOFMEMORY;
			newentry->playernum=operating_player;
			newentry->command=CMD_RCV_MOVE;
			newentry->op2=0;

			if (token[0]=='N')
				newentry->op1=DIRECTION_N;
			if (token[0]=='S')
				newentry->op1=DIRECTION_S;
			if (token[0]=='E')
				newentry->op1=DIRECTION_E;
			if (token[0]=='W')
				newentry->op1=DIRECTION_W;

			newentry->next=NULL;

			/* Add to end of list */
			if (current==NULL)
				head=current=newentry;
			else
			{
				current->next=newentry;
				current=newentry;
			}


			break;
		}


		token=strtok(NULL," ");

	}

	*rlist=head;


	return 0;
}





int proto_sendfullbuffer(char *buffer,int blen)
{
	char *real_ptr;
	int real_len,sent_len;

	real_ptr=buffer;
	real_len=blen;

	while (real_len>0)
	{
		sent_len=net_sendblock(real_ptr,real_len);
		if (FAILURE(sent_len))
		{
			DEBUG_PRINT("Proto error on sendfullbuffer.\n");
			return sent_len;
		}
		real_ptr+=sent_len;
		real_len-=sent_len;
	}

#ifdef DEBUG_PROTO
	DEBUG_PRINT("Sent:");
	DEBUG_PRINT(buffer);
#endif
	return 0;
}

/*
 *  Read until a \n.  Discard the \n.  Null terminate.
 *
 *
 *
 */

int proto_receiveline()
{
	int rval;

	/* Simplistic approach.  Always start at the beginning of the buffer. */

	DEBUG_PRINT("Waiting for data..\n");

	readbuffer_offset=0;

	if (readbuffer==NULL)
		proto_enlargebuffer();

	while (1)
	{
		rval=net_getblock(readbuffer+readbuffer_offset,1);

		if (rval<0)
		{
			DEBUG_PRINT("Network error on net_getblock().\n");
			return rval;
		}
		if (rval>0 && readbuffer[readbuffer_offset]=='\n')
		{
			readbuffer[readbuffer_offset]=0x00;
#ifdef DEBUG_PROTO
			DEBUG_PRINT("Received:");
			DEBUG_PRINT(readbuffer);
			DEBUG_PRINT("\n");
#endif
			return 0;
		}
		readbuffer_offset+=rval;

		if (readbuffer_offset>=readbuffer_size)
		{
			rval=proto_enlargebuffer();
			if (FAILURE(rval))
			{
				DEBUG_PRINT("Proto error on proto_enlargebuffer().\n");
				return rval;
			}

		}
	}


	return ERR_UNREACHABLE;
}



int proto_enlargebuffer()
{
	char *newbuffer;

	newbuffer=malloc(readbuffer_size+BUFFER_STEP_SIZE);

	if (newbuffer==NULL)
		return ERR_OUTOFMEMORY;

	if (readbuffer!=NULL)
	{
		memcpy(newbuffer,readbuffer,readbuffer_size);
		free(readbuffer);
	}
	readbuffer=newbuffer;
	readbuffer_size+=BUFFER_STEP_SIZE;

	return 0;
}


